import multiprocessing as mp
import os
import sys
from functools import partial
from typing import Tuple

import click
import pandas as pd
import xarray as xr
import yaml

from src.mhw_detect.detection.detect import prepare_data
from src.mhw_detect.detection.parser import (
    check_climato_period,
    check_file_exist,
    count_files,
    parse_data,
    parse_param,
)


def check_extension(ctx, param, value):
    if not value.endswith((".parquet", ".csv")):
        raise click.BadParameter("The extension must be .parquet or .csv")
    else:
        return value


@click.command(
    help="""
    Detect extreme events
    """
)
@click.option(
    "--config",
    "-c",
    type=click.Path(exists=True),
    required=True,
    help="Specify configuration file",
)
@click.option(
    "--geographical-subset",
    "-g",
    type=(
        click.FloatRange(min=-90, max=90),
        click.FloatRange(min=-90, max=90),
        click.FloatRange(min=-180, max=180),
        click.FloatRange(min=-180, max=180),
    ),
    help="The geographical subset as "
    + "minimal latitude, maximal latitude, "
    + "minimal longitude and maximal longitude.\n\n"
    + "If set, the detection will be done on the subsetted global dataset given in \
            config file (not the cuts) and sequentially.",
)
@click.option(
    "--categ-map",
    type=str,
    help="Generate category map in a netcdf file.",
    show_default=True,
)
@click.option(
    "--output-df",
    type=click.UNPROCESSED,
    help="Give a name to the output dataframe.\
            Two extensions are available: csv and parquet (default).\
            Save in csv if you want to open the dataframe with excel.\
            Parquet is more efficient and takes less disk space.",
    default="data.parquet",
    show_default=True,
    callback=check_extension,
)
def extreme_events(
    config: str,
    geographical_subset: Tuple[float, float, float, float],
    categ_map: str,
    output_df: str,
):
    mask = bool(categ_map)

    conf = yaml.safe_load(open(config))

    output = conf["output_detection"]

    try:
        check_file_exist(conf)
    except Exception as error:
        click.echo(repr(error))
        sys.exit()

    param = parse_param(conf)

    def output_path(file_name):
        return os.path.join(output, file_name)

    def save_dataframe(dataframe):
        if output_df.endswith(".parquet"):
            dataframe.to_parquet(output_path(output_df))
        else:
            dataframe.to_csv(output_path(output_df), sep=";")

    if geographical_subset is not None:
        data = parse_data(conf, False)

        lat = geographical_subset[0:2]
        lon = geographical_subset[2:4]

        prepare_data(
            0, output, mask=mask, lat=lat, lon=lon, p=param["pctile"], **data, **param
        )

        click.echo("Creating csv")
        df = pd.read_csv(output_path("0.txt"), sep=";")
        save_dataframe(df)

        if categ_map:
            os.rename(output_path("0.txt.nc"), output_path(categ_map))

    else:
        data = parse_data(conf)
        nfile = count_files(conf)

        if "clim" not in data:
            check_climato_period(conf)

        pool = mp.Pool()
        pool.imap(
            partial(
                prepare_data,
                outdir=output,
                mask=mask,
                p=param["pctile"],
                **data,
                **param
            ),
            range(1, nfile),
        )
        pool.close()
        pool.join()

        click.echo("Computation done")

        click.echo("Saving dataframe")

        def f(i):
            return pd.read_csv(i, sep=";")

        filepaths = [output_path(str(i) + ".txt") for i in range(1, nfile)]
        df = pd.concat(map(f, filepaths))
        save_dataframe(df)

        if categ_map:
            click.echo("Creating mask")
            mask = xr.open_mfdataset(output_path("*.txt.nc"))
            comp = dict(zlib=True)
            encoding = {var: comp for var in mask.data_vars}
            mask.to_netcdf(output_path(categ_map), encoding=encoding)
            p = [output_path(str(g) + ".txt.nc") for g in range(1, nfile)]

            for path in p:
                try:
                    os.remove(path)
                except OSError:
                    click.echo("Error while deleting file: ", path)


if __name__ == "__main__":
    extreme_events()



import os
import xarray as xr
import itertools
import click
import yaml
from dask.diagnostics import ProgressBar


# Créer un dataset par chunk dask
def split_by_chunks(dataset):
    chunk_slices = {}
    for dim, chunks in dataset.chunks.items():
        slices = []
        start = 0
        for chunk in chunks:
            if start >= dataset.sizes[dim]:
                break
            stop = start + chunk
            slices.append(slice(start, stop))
            start = stop
        chunk_slices[dim] = slices
    for slices in itertools.product(*chunk_slices.values()):
        selection = dict(zip(chunk_slices.keys(), slices))
        yield dataset[selection]


@click.command(
    help="""
    Cut datasets in multiple files
    """
)
@click.option(
    "--config", "-c", type=str, required=True, help="Specify configuration file"
)
def cut(config: str):
    conf = yaml.safe_load(open(config, "r"))

    lat = conf["cut"]["nb_lat"]
    lon = conf["cut"]["nb_lon"]

    for datas in conf["data"]:

        path = conf["data"][datas]["path"]
        ds = xr.open_dataset(path)

        if "latitude" in ds.coords:
            chunks = {"latitude": lat, "longitude": lon}
        else:
            chunks = {"lat": lat, "lon": lon}

        ds = xr.open_dataset(path, chunks=chunks)

        chunked_ds = list(split_by_chunks(ds))

        p = [
            os.path.dirname(path) + "/Cut_" + str(i + 1) + ".nc"
            for i in range(len(chunked_ds))
        ]
        print("Saving in " + os.path.dirname(path))
        delayed = xr.save_mfdataset(datasets=chunked_ds, paths=p, compute=False)
        with ProgressBar():
            delayed.compute()
        print(str(len(chunked_ds)) + " files saved.")


if __name__ == "__main__":
    cut()



import numpy as np
import numba as nb
from numba import njit


@njit(
    nb.types.Tuple((nb.float64[:], nb.float64[:]))(
        nb.int64,
        nb.int64,
        nb.int64,
        nb.float64[:],
        nb.float64[:],
        nb.int64,
        nb.float32[:],
        nb.float64[:],
        nb.int64,
        nb.int64,
        nb.int64,
    ),
    cache=True,
)
def threshAndClim(
    lenClimYear,
    feb29,
    pctile,
    thresh_climYear,
    seas_climYear,
    TClim,
    tempClim,
    doyClim,
    clim_start,
    clim_end,
    windowHalfWidth,
):
    """

    Compute climatology mean and threshold.

    Written by Eric Oliver, Institue for Marine and Antarctic Studies, University of Tasmania, Feb-Mar 2015
    Optimised by Jonathan Brouillet, Mercator Ocean International, Toulouse, France, 2021

    """

    for d in range(1, lenClimYear + 1):
        # Special case for Feb 29
        if d != feb29:
            # find all indices for each day of the year +/- windowHalfWidth and from them calculate the threshold
            tt0 = np.where(doyClim[clim_start : clim_end + 1] == d)[0]
            # If this doy value does not exist (i.e. in 360-day calendars) then skip it
            if len(tt0) != 0:
                tt = np.empty(0)
                for w in range(-windowHalfWidth, windowHalfWidth + 1):
                    tt = np.append(tt, clim_start + tt0 + w)

                tt = tt[tt >= 0]  # Reject indices "before" the first element
                tt = tt[tt < TClim]  # Reject indices "after" the last element
                thresh_climYear[d - 1] = np.nanpercentile(
                    tempClim[tt.astype(np.int64)], pctile
                )
                seas_climYear[d - 1] = np.nanmean(tempClim[tt.astype(np.int64)])

    return thresh_climYear, seas_climYear


@njit(cache=True)
def mhwproperties(t, temp, clim, n, mhw):
    """

    Add mhw characteristics in Numba dictionnary.

    Written by Eric Oliver, Institue for Marine and Antarctic Studies, University of Tasmania, Feb-Mar 2015
    Optimised by Jonathan Brouillet, Mercator Ocean International, Toulouse, France, 2021

    """

    T = len(t)
    categories = np.arange(1, 5, dtype=np.float64)
    for ev in range(n):
        tt_start = np.where(t == mhw["time_start"][ev])[0][0]
        tt_end = np.where(t == mhw["time_end"][ev])[0][0]
        appendjit(mhw, "index_start", tt_start)
        appendjit(mhw, "index_end", tt_start)
        temp_mhw = temp[tt_start : tt_end + 1]
        thresh_mhw = clim["thresh"][tt_start : tt_end + 1]
        seas_mhw = clim["seas"][tt_start : tt_end + 1]
        mhw_relSeas = temp_mhw - seas_mhw
        mhw_relThresh = temp_mhw - thresh_mhw
        mhw_relThreshNorm = (temp_mhw - thresh_mhw) / (thresh_mhw - seas_mhw)
        mhw_abs = temp_mhw
        # Find peak
        tt_peak = np.argmax(mhw_relSeas)
        appendjit(mhw, "time_peak", mhw["time_start"][ev] + tt_peak)
        appendjit(mhw, "index_peak", tt_start + tt_peak)
        # MHW Duration
        appendjit(mhw, "duration", len(mhw_relSeas))
        # MHW Intensity metrics
        appendjit(mhw, "intensity_max", mhw_relSeas[tt_peak])
        appendjit(mhw, "intensity_mean", mhw_relSeas.mean())
        appendjit(mhw, "intensity_var", np.sqrt(mhw_relSeas.var()))
        appendjit(mhw, "intensity_cumulative", mhw_relSeas.sum())
        appendjit(mhw, "intensity_max_relThresh", mhw_relThresh[tt_peak])
        appendjit(mhw, "intensity_mean_relThresh", mhw_relThresh.mean())
        appendjit(mhw, "intensity_var_relThresh", np.sqrt(mhw_relThresh.var()))
        appendjit(mhw, "intensity_cumulative_relThresh", mhw_relThresh.sum())
        appendjit(mhw, "intensity_max_abs", mhw_abs[tt_peak])
        appendjit(mhw, "intensity_mean_abs", mhw_abs.mean())
        appendjit(mhw, "intensity_var_abs", np.sqrt(mhw_abs.var()))
        appendjit(mhw, "intensity_cumulative_abs", mhw_abs.sum())
        # Fix categories
        tt_peakCat = np.argmax(mhw_relThreshNorm)
        cats = np.floor(1.0 + mhw_relThreshNorm)
        appendjit(
            mhw,
            "category",
            categories[np.int64(np.min(np.array([cats[tt_peakCat], 4]))) - 1],
        )
        appendjit(mhw, "duration_moderate", np.sum(cats == 1.0))
        appendjit(mhw, "duration_strong", np.sum(cats == 2.0))
        appendjit(mhw, "duration_severe", np.sum(cats == 3.0))
        appendjit(mhw, "duration_extreme", np.sum(cats >= 4.0))

        # Rates of onset and decline
        # Requires getting MHW strength at "start" and "end" of event (continuous: assume start/end half-day before/after first/last point)
        try:
            if tt_start > 0:
                mhw_relSeas_start = 0.5 * (
                    mhw_relSeas[0] + temp[tt_start - 1] - clim["seas"][tt_start - 1]
                )
                appendjit(
                    mhw,
                    "rate_onset",
                    (mhw_relSeas[tt_peak] - mhw_relSeas_start)
                    / np.float64(tt_peak + 0.5),
                )
            else:  # MHW starts at beginning of time series
                if (
                    tt_peak == 0
                ):  # Peak is also at begining of time series, assume onset time = 1 day
                    appendjit(
                        mhw, "rate_onset", (mhw_relSeas[tt_peak] - mhw_relSeas[0]) / 1.0
                    )
                else:
                    appendjit(
                        mhw,
                        "rate_onset",
                        (mhw_relSeas[tt_peak] - mhw_relSeas[0]) / tt_peak,
                    )
        except:
            appendjit(mhw, "rate_onset", np.nan)
            
        try:
            if tt_end < T - 1:
                mhw_relSeas_end = 0.5 * (
                    mhw_relSeas[-1] + temp[tt_end + 1] - clim["seas"][tt_end + 1]
                )
                appendjit(
                    mhw,
                    "rate_decline",
                    (mhw_relSeas[tt_peak] - mhw_relSeas_end)
                    / np.float64(tt_end - tt_start - tt_peak + 0.5),
                )
            else:  # MHW finishes at end of time series
                if (
                    tt_peak == T - 1
                ):  # Peak is also at end of time series, assume decline time = 1 day
                    appendjit(
                        mhw,
                        "rate_decline",
                        (mhw_relSeas[tt_peak] - mhw_relSeas[-1]) / 1.0,
                    )
                else:
                    appendjit(
                        mhw,
                        "rate_decline",
                        (mhw_relSeas[tt_peak] - mhw_relSeas[-1])
                        / np.float64(tt_end - tt_start - tt_peak),
                    )
        except:
            appendjit(mhw, "rate_decline", np.nan)

    return tt_peak


@njit(cache=True)
def appendjit(mhw, key, val):
    """

    Numba-friendly append function.

    Written by Jonathan Brouillet, Mercator Ocean International, Toulouse, France, 2021

    """
    mhw[key] = np.append(mhw[key], np.float32(val))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cut', 'cut..ipynb_checkpoints', 'detection', 'detection..ipynb_checkpoints']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.1.3',
 'dask>=2021.10.0',
 'netCDF4>=1.5.7',
 'numba>=0.54.1',
 'numpy>=1.20.3',
 'pandas>=1.3.5',
 'pyarrow>=11.0.0,<12.0.0',
 'scipy>=1.7.3',
 'xarray>=2022.6.0']

entry_points = \
{'console_scripts': ['mhw-cut = src.mhw_detect.cut.cut:cut',
                     'mhw-detect = '
                     'src.mhw_detect.detection.mhw:extreme_events']}

setup_kwargs = {
    'name': 'mhw-detect',
    'version': '0.2.0',
    'description': 'Fast marine heatwaves and extrem events detector based on https://github.com/ecjoliver/marineHeatWaves',
    'long_description': "# MHW Detector\n\nMarine heatwaves detector based on https://github.com/ecjoliver/marineHeatWaves.\n\nThis package integrates a numba optimised version of ecjoliver's implementation for MHW detection with multiprocessing capabilities to compute detection over every coordinates of the dataset.\n\nThis code is not only for detecting MHW. It can also be used to detect extrem events of any variables like chla, pH, O2, etc ...\n\n## Installation\n> pip install mhw-detect\n\n\n### Dependencies\n- xarray\n- numba\n- scipy\n- dask\n- numpy\n- pandas\n- netcdf4\n- click\n- pyarrow\n\n## Usage\n### Configuration file\nWith mhw-detect no need for kilometers of parameters in command line. You just need to write a configuration file in which you put every parameters like an identity card of your next detection.\n\n```yaml\ndata:\n  data :\n    path : '/folder/sst.nc'\n    var : 'sst'\n  clim :\n    path : '/folder_clim/clim.nc'\n    var : 'sst'\n  percent :\n    path : '/folder_percent/percentile.nc'\n    var : 'sst'\n\nparams:\n    depth : 0\n    climatologyPeriod : [null, null]\n    pctile : 90\n    windowHalfWidth : 5\n    smoothPercentile : True\n    smoothPercentileWidth : 31\n    minDuration : 5\n    joinAcrossGaps : True\n    maxGap : 2\n    maxPadLength: False\n    coldSpells : False\n    Ly : False\n\ncut:\n  nb_lat : 157\n  nb_lon : 72\n\noutput_detection : '/my/path/to/folder_result/'\n```\n\n- `data` : specifies the paths and variables you want to use. Do not specify clim and percent if you want them to be computed during the detection.\n- `params` : specifies the parameters of the detection. See section below.\n- `cut` : specifies the number of latitude and longitude for geospatial dataset cutting.\n- `output_detection` : specifies the folder in which to save the results.\n\n\n### Detection parameters\nFrom https://github.com/ecjoliver/marineHeatWaves.\n\n```\nclimatologyPeriod      Period over which climatology is calculated, specified\n                        as list of start and end years. Default ([null, null]) is to calculate\n                        over the full range of years in the supplied time series.\n                        Alternate periods suppled as a list e.g. [1983,2012].\n                        Unused if precalculated clim and percentile are set.\n\npctile                 Threshold percentile (%) for detection of extreme values\n                        (DEFAULT = 90)\n\nwindowHalfWidth        Width of window (one sided) about day-of-year used for\n                        the pooling of values and calculation of threshold percentile\n                        (DEFAULT = 5 [days])\n\nsmoothPercentile       Boolean switch indicating whether to smooth the threshold\n                        percentile timeseries with a moving average (DEFAULT = True)\n\nsmoothPercentileWidth  Width of moving average window for smoothing threshold\n                        (DEFAULT = 31 [days])\n\nminDuration            Minimum duration for acceptance detected MHWs\n                        (DEFAULT = 5 [days])\n\njoinAcrossGaps         Boolean switch indicating whether to join MHWs\n                        which occur before/after a short gap (DEFAULT = True)\n\nmaxGap                 Maximum length of gap allowed for the joining of MHWs\n                        (DEFAULT = 2 [days])\n\nmaxPadLength           Specifies the maximum length [days] over which to interpolate\n                        (pad) missing data (specified as nans) in input temp time series.\n                        i.e., any consecutive blocks of NaNs with length greater\n                        than maxPadLength will be left as NaN. Set as an integer.\n                        (DEFAULT = False, interpolates over all missing values).\n\ncoldSpells             Specifies if the code should detect cold events instead of\n                        heat events. (DEFAULT = False)\n```\n\n### Datasets coordinates\nAll datasets must have lat/latitude, lon/longitude and time as coordinates. `depth` coordinate is allowed for the main dataset. Currently, the depth as to be specified via its index in the coordinate array. Giving directly the wanted depth will be added later.\n\nThe percentile dataset must have a `quantile` coordinate as a dimension for the variable. It is useful in the case you want to do the detection with different quantile (90, 99).\n\n### Step 1 : Geospatial cutting (optionnal but recommended)\nTo use multiprocessing efficiently, the datasets must be cut in several smaller datasets over the lat/lon dimensions. Call `mhw-cut` with your config file to make it happen. Each sub-datasets will be called Cut_X.nc where X is the number of the cut (that is why your datasets (data, clim, percentile) must be in different folders).\n\nThe number of cuts does not matter, chunk size does. To find suitables nb_lat and nb_lon, it is better to use a notebook.\n>import xarray as xr\nds = xr.open_dataset('dataset.nc', chunks={'latitude': nb_lat, 'longitude': nb_lon})\nds\n\nnb_lat and nb_lon should be multiples of the latitude and longitude dimensions and choose carefully to have chunks of size over 10Mb (see Dask documentation for more details). Printing ds on a notebook gives you the size of the chunk (cut).\n\nPlease note that this step will double the space used in your disk.\n\n\n### Step 2 : Detection\nCall `mhw-detect` to detect MHW. With multiprocessing, each cut is processed in parallel. For a cut, you will get in `output_detection` a text file with the results of the detection. Finally, when all the detections are done, every text files are concatenated into one csv (with ; as a separator).\n\nIf you do not want to use multiprocessing just to make a detection on a small geospatial subset, you can give the option `-g lat_min lat_max lon_min lon_max` to the command.\n\n\n### Commands\n#### Geospatial cut\n```\nUsage: mhw-cut [OPTIONS]\n\n  Cut datasets in multiple files\n\nOptions:\n  -c, --config PATH  Specify configuration file  [required]\n  --help             Show this message and exit.\n```\n\n#### Detection\n```\nUsage: mhw-detect [OPTIONS]\n\n  Detect extreme events\n\nOptions:\n  -c, --config PATH               Specify configuration file  [required]\n\n  -g, --geographical-subset <FLOAT RANGE FLOAT RANGE FLOAT RANGE FLOAT RANGE>...\n                                  The geographical subset as minimal latitude,\n                                  maximal latitude, minimal longitude and\n                                  maximal longitude.\n\n                                  If set, the detection will be done on the\n                                  subsetted global dataset given in config\n                                  file (not the cuts) and sequentially.\n\n  --categ-map TEXT                Generate category map in a netcdf file.\n\n  --output-df TEXT                Give a name to the output dataframe. Two\n                                  extensions are available: csv and parquet\n                                  (default).     Save in csv if you want to\n                                  open the dataframe with excel. Parquet is\n                                  more efficient and takes less disk space.\n                                  [default: data.parquet]\n\n  --help                          Show this message and exit.\n  ```\n\n\n## Output\nHere is an example of the output csv file. Every detected MHW are listed.\n|     |      lat     |     lon     |   time_deb   |   time_end   |   time_peak  |  duration  |  duration_mod  |  duration_str  |  duration_sev  |  duration_ext  |  categ  |     imax    |     imean     |     ivar    |  rate_onset  |  rate_decline |\n|:---:|:------------:|:-----------:|:------------:|:------------:|:------------:|:----------:|:--------------:|:--------------:|:--------------:|:--------------:|:-------:|:-----------:|:-------------:|:-----------:|:------------:|:-------------:|\n|  0  |  -76.916664  |  -180.0     |  2003-01-01  |  2003-01-18  |  2003-01-04  |  18        |  5.0           |  2.0           |  5.0           |  6.0           |  4.0    |  2.341543   |  1.415663     |  0.551971   |  0.1867      |  0.2049       |\n|  1  |  -76.916664  |  -180.0     |  2003-04-10  |  2003-10-23  |  2003-04-13  |  197       |  0.0           |  0.0           |  0.0           |  0.0           |  4.0    |  4.2300e-8  |  -4.8800e-10  |  7.0377e-9  |  1.5032e-8   |  2.7109e-10   |\n|  2  |  -76.916664  |  -180.0     |  2003-12-18  |  2003-12-23  |  2003-12-21  |  6         |  0.0           |  1.0           |  2.0           |  3.0           |  4.0    |  2.325211   |  1.858383     |  0.367969   |  0.482987    |  0.613132     |\n|  3  |  -76.916664  |  -179.9166  |  2003-01-01  |  2003-01-18  |  2003-01-04  |  18        |  5.0           |  2.0           |  5.0           |  6.0           |  4.0    |  2.327172   |  1.420817     |  0.544248   |  0.182604    |  0.203315     |\n\n## To do\n- Add shapefile usage.\n- Add an option to remove text files.\n\n## Contribution\n- Install [poetry](https://python-poetry.org/) in your python/conda environment.\n- Clone this repository and checkout to your branch.\n- Create a poetry environment for dev by running `poetry install`.\n- Make your dev.\n- Test your dev with the command `poetry run python monscript.py ...`.\n- Commit and push.\n- Ask for a merge request.\n\n\n\n## References\n\nHobday, A.J. et al. (2016), A hierarchical approach to defining marine heatwaves, Progress in Oceanography, 141, pp. 227-238, doi: 10.1016/j.pocean.2015.12.014 [pdf](http://passage.phys.ocean.dal.ca/~olivere/docs/Hobdayetal_2016_PO_HierarchMHWDefn.pdf)\n",
    'author': 'John Brouillet',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.12',
}


setup(**setup_kwargs)



import os
import glob
import xarray as xr


def check_precalc_clim_thresh(data):
    if not {"clim", "percent"} <= data.keys():
        print(
            "Precalculated mean and/or percentile have not been referenced in config file. \n\
        They will be recalculated by Hobday's method."
        )


def parse_data(conf, cut=True):
    if cut:
        parse_path = lambda p: os.path.join(os.path.dirname(p), "Cut_")
    else:
        parse_path = lambda p: p

    data = {
        datas: [parse_path(conf["data"][datas]["path"]), conf["data"][datas]["var"]]
        for datas in conf["data"]
    }
    check_precalc_clim_thresh(data)
    return data


def parse_param(conf):
    return {params: conf["params"][params] for params in conf["params"]}


def count_files(conf):
    return (
        len(
            [
                name
                for name in glob.glob(
                    os.path.dirname(conf["data"]["data"]["path"]) + "/Cut_*.nc"
                )
            ]
        )
        + 1
    )


def check_climato_period(conf):
    ds = xr.open_dataset(conf["data"]["data"]["path"])
    period = conf["params"]["climatologyPeriod"]

    min = ds.time.min().values.astype("datetime64[Y]").astype(int) + 1970
    max = ds.time.max().values.astype("datetime64[Y]").astype(int) + 1970

    assert (period[0] >= min) or (
        period[1] <= max
    ), "The dataset does not cover the climatologyPeriod."
